/*
Copyright (c) 2019-2022 Digital Energy Cloud Solutions LLC. All Rights Reserved.
Authors:
Petr Krutov, <petr.krutov@digitalenergy.online>
Stanislav Solovev, <spsolovev@digitalenergy.online>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
Terraform DECORT provider - manage resources provided by DECORT (Digital Energy Cloud
Orchestration Technology) with Terraform by Hashicorp.

Source code: https://repository.basistech.ru/BASIS/terraform-provider-decort

Please see README.md to learn where to place source code so that it
builds seamlessly.

Documentation: https://repository.basistech.ru/BASIS/terraform-provider-decort/wiki
*/

package disks

import (
	"context"
	"fmt"
	"strconv"

	"repository.basistech.ru/BASIS/decort-golang-sdk/pkg/cloudbroker/disks"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/constants"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/controller"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/dc"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/status"

	"github.com/hashicorp/terraform-plugin-sdk/v2/diag"
	"github.com/hashicorp/terraform-plugin-sdk/v2/helper/schema"
	"github.com/hashicorp/terraform-plugin-sdk/v2/helper/validation"
	log "github.com/sirupsen/logrus"
)

func resourceDiskCreate(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	c := m.(*controller.ControllerCfg)

	err := checkParamsExists(ctx, d, m)
	if err != nil {
		return diag.FromErr(err)
	}

	req := disks.CreateRequest{
		AccountID: uint64(d.Get("account_id").(int)),
		GID:       uint64(d.Get("gid").(int)),
		Name:      d.Get("disk_name").(string),
		Size:      uint64(d.Get("size_max").(int)),
		Type:      d.Get("type").(string),
	}

	if desc, ok := d.GetOk("desc"); ok {
		req.Description = desc.(string)
	}

	if ssdSize, ok := d.GetOk("ssd_size"); ok {
		req.SSDSize = uint64(ssdSize.(int))
	}

	if sepID, ok := d.GetOk("sep_id"); ok {
		req.SEPID = uint64(sepID.(int))
	}

	if pool, ok := d.GetOk("pool"); ok {
		req.Pool = pool.(string)
	}

	diskID, err := c.CloudBroker().Disks().Create(ctx, req)
	if err != nil {
		return diag.FromErr(err)
	}

	d.SetId(strconv.FormatUint(diskID, 10))

	w := dc.Warnings{}

	if iotuneRaw, ok := d.GetOk("iotune"); ok {
		iot := iotuneRaw.([]interface{})[0]
		iotune := iot.(map[string]interface{})
		req := disks.LimitIORequest{
			DiskID:           diskID,
			IOPS:             uint64(iotune["total_iops_sec"].(int)),
			ReadBytesSec:     uint64(iotune["read_bytes_sec"].(int)),
			ReadBytesSecMax:  uint64(iotune["read_bytes_sec_max"].(int)),
			ReadIOPSSec:      uint64(iotune["read_iops_sec"].(int)),
			ReadIOPSSecMax:   uint64(iotune["read_iops_sec_max"].(int)),
			SizeIOPSSec:      uint64(iotune["size_iops_sec"].(int)),
			TotalBytesSec:    uint64(iotune["total_bytes_sec"].(int)),
			TotalBytesSecMax: uint64(iotune["total_bytes_sec_max"].(int)),
			TotalIOPSSecMax:  uint64(iotune["total_iops_sec_max"].(int)),
			TotalIOPSSec:     uint64(iotune["total_iops_sec"].(int)),
			WriteBytesSec:    uint64(iotune["write_bytes_sec"].(int)),
			WriteBytesSecMax: uint64(iotune["write_bytes_sec_max"].(int)),
			WriteIOPSSec:     uint64(iotune["write_iops_sec"].(int)),
			WriteIOPSSecMax:  uint64(iotune["write_iops_sec_max"].(int)),
		}

		_, err := c.CloudBroker().Disks().LimitIO(ctx, req)
		if err != nil {
			w.Add(err)
		}
	}

	if shareable := d.Get("shareable"); shareable.(bool) {
		_, err := c.CloudBroker().Disks().Share(ctx, disks.ShareRequest{
			DiskID: diskID,
		})
		if err != nil {
			w.Add(err)
		}
	}

	return append(w.Get(), resourceDiskRead(ctx, d, m)...)
}

func resourceDiskRead(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	c := m.(*controller.ControllerCfg)
	w := dc.Warnings{}

	disk, err := utilityDiskCheckPresence(ctx, d, m)
	if err != nil {
		d.SetId("")
		return diag.FromErr(err)
	}

	hasChangeState := false

	switch disk.Status {
	case status.Destroyed, status.Purged:
		d.Set("disk_id", 0)
		d.SetId("")
		return resourceDiskCreate(ctx, d, m)
	case status.Deleted:
		hasChangeState = true
		req := disks.RestoreRequest{
			DiskID: disk.ID,
		}

		if reason, ok := d.GetOk("reason"); ok {
			req.Reason = reason.(string)
		} else {
			req.Reason = "Terraform automatic restore"
		}

		_, err := c.CloudBroker().Disks().Restore(ctx, req)
		if err != nil {
			w.Add(err)
		}
	case status.Assigned:
	case status.Modeled:
		return diag.Errorf("The disk is in status: %s, please, contact support for more information", disk.Status)
	case status.Creating:
	case status.Created:
	case status.Allocated:
	case status.Unallocated:
	}

	if hasChangeState {
		disk, err = utilityDiskCheckPresence(ctx, d, m)
		if err != nil {
			return diag.FromErr(err)
		}
	}

	flattenDisk(d, disk)

	return w.Get()
}

func resourceDiskUpdate(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	c := m.(*controller.ControllerCfg)
	w := dc.Warnings{}

	err := checkParamsExists(ctx, d, m)
	if err != nil {
		return diag.FromErr(err)
	}

	disk, err := utilityDiskCheckPresence(ctx, d, m)
	if err != nil {
		d.SetId("")
		return diag.FromErr(err)
	}

	hasChangeState := false

	switch disk.Status {
	case status.Destroyed, status.Purged:
		d.Set("disk_id", 0)
		d.SetId("")
		return resourceDiskCreate(ctx, d, m)
	case status.Deleted:
		hasChangeState = true
		req := disks.RestoreRequest{
			DiskID: disk.ID,
		}

		if reason, ok := d.GetOk("reason"); ok {
			req.Reason = reason.(string)
		} else {
			req.Reason = "Terraform automatic restore"
		}

		_, err := c.CloudBroker().Disks().Restore(ctx, req)
		if err != nil {
			return diag.FromErr(err)
		}
	case status.Assigned:
	case status.Modeled:
		return diag.Errorf("The disk is in status: %s, please, contact support for more information", disk.Status)
	case status.Creating:
	case status.Created:
	case status.Allocated:
	case status.Unallocated:
	}

	if hasChangeState {
		disk, err = utilityDiskCheckPresence(ctx, d, m)
		if err != nil {
			d.SetId("")
			return diag.FromErr(err)
		}
	}

	if d.HasChange("size_max") {
		oldSize, newSize := d.GetChange("size_max")
		if oldSize.(int) < newSize.(int) {
			log.Debugf("resourceDiskUpdate: resizing disk ID %s - %d GB -> %d GB",
				d.Id(), oldSize.(int), newSize.(int))

			_, err := c.CloudBroker().Disks().Resize(ctx, disks.ResizeRequest{
				DiskID: disk.ID,
				Size:   uint64(newSize.(int)),
			})
			if err != nil {
				w.Add(err)
			}
		} else if oldSize.(int) > newSize.(int) {
			return diag.FromErr(fmt.Errorf("resourceDiskUpdate: Disk ID %s - reducing disk size is not allowed", d.Id()))
		}
	}

	if d.HasChange("disk_name") {
		_, err := c.CloudBroker().Disks().Rename(ctx, disks.RenameRequest{
			DiskID: disk.ID,
			Name:   d.Get("disk_name").(string),
		})
		if err != nil {
			w.Add(err)
		}
	}

	if d.HasChange("iotune") {
		iot := d.Get("iotune").([]interface{})[0]
		iotune := iot.(map[string]interface{})
		req := disks.LimitIORequest{
			IOPS:             uint64(iotune["total_iops_sec"].(int)),
			ReadBytesSec:     uint64(iotune["read_bytes_sec"].(int)),
			ReadBytesSecMax:  uint64(iotune["read_bytes_sec_max"].(int)),
			ReadIOPSSec:      uint64(iotune["read_iops_sec"].(int)),
			ReadIOPSSecMax:   uint64(iotune["read_iops_sec_max"].(int)),
			SizeIOPSSec:      uint64(iotune["size_iops_sec"].(int)),
			TotalBytesSec:    uint64(iotune["total_bytes_sec"].(int)),
			TotalBytesSecMax: uint64(iotune["total_bytes_sec_max"].(int)),
			TotalIOPSSecMax:  uint64(iotune["total_iops_sec_max"].(int)),
			TotalIOPSSec:     uint64(iotune["total_iops_sec"].(int)),
			WriteBytesSec:    uint64(iotune["write_bytes_sec"].(int)),
			WriteBytesSecMax: uint64(iotune["write_bytes_sec_max"].(int)),
			WriteIOPSSec:     uint64(iotune["write_iops_sec"].(int)),
			WriteIOPSSecMax:  uint64(iotune["write_iops_sec_max"].(int)),
		}

		_, err := c.CloudBroker().Disks().LimitIO(ctx, req)
		if err != nil {
			w.Add(err)
		}
	}

	if d.HasChange("shareable") {
		old, new := d.GetChange("shareable")
		if !old.(bool) && new.(bool) && !disk.Shareable {
			_, err := c.CloudBroker().Disks().Share(ctx, disks.ShareRequest{
				DiskID: disk.ID,
			})
			if err != nil {
				w.Add(err)
			}
		}
		if old.(bool) && !new.(bool) && disk.Shareable {
			_, err := c.CloudBroker().Disks().Unshare(ctx, disks.UnshareRequest{
				DiskID: disk.ID,
			})
			if err != nil {
				w.Add(err)
			}
		}
	}

	return append(w.Get(), resourceDiskRead(ctx, d, m)...)
}

func resourceDiskDelete(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	c := m.(*controller.ControllerCfg)

	disk, err := utilityDiskCheckPresence(ctx, d, m)
	if err != nil {
		d.SetId("")
		return diag.FromErr(err)
	}

	req := disks.DeleteRequest{
		DiskID:      disk.ID,
		Detach:      d.Get("detach").(bool),
		Permanently: d.Get("permanently").(bool),
		Reason:      d.Get("reason").(string),
	}

	_, err = c.CloudBroker().Disks().Delete(ctx, req)
	if err != nil {
		return diag.FromErr(err)
	}

	d.SetId("")

	return nil
}

func resourceDiskSchemaMake() map[string]*schema.Schema {
	rets := map[string]*schema.Schema{
		"account_id": {
			Type:     schema.TypeInt,
			Required: true,
			ForceNew: true,
		},
		"gid": {
			Type:     schema.TypeInt,
			Required: true,
			ForceNew: true,
		},
		"disk_name": {
			Type:     schema.TypeString,
			Required: true,
		},
		"type": {
			Type:         schema.TypeString,
			Required:     true,
			ValidateFunc: validation.StringInSlice([]string{"D", "B", "T"}, false),
			Description:  "The type of disk in terms of its role in compute: 'B=Boot, D=Data, T=Temp'",
		},
		"desc": {
			Type:     schema.TypeString,
			Optional: true,
			Computed: true,
		},
		"size_max": {
			Type:     schema.TypeInt,
			Required: true,
		},
		"ssd_size": {
			Type:     schema.TypeInt,
			Optional: true,
		},
		"sep_id": {
			Type:     schema.TypeInt,
			Optional: true,
			Computed: true,
		},
		"pool": {
			Type:     schema.TypeString,
			Optional: true,
			Computed: true,
		},
		"detach": {
			Type:        schema.TypeBool,
			Optional:    true,
			Default:     false,
			Description: "detach disk from machine first",
		},
		"permanently": {
			Type:        schema.TypeBool,
			Optional:    true,
			Default:     false,
			Description: "whether to completely delete the disk, works only with non attached disks",
		},
		"reason": {
			Type:        schema.TypeString,
			Optional:    true,
			Default:     "",
			Description: "reason for an action",
		},
		"shareable": {
			Type:     schema.TypeBool,
			Optional: true,
			Computed: true,
		},
		"restore": {
			Type:        schema.TypeBool,
			Optional:    true,
			Default:     false,
			Description: "restore deleting disk",
		},
		"account_name": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"acl": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"boot_partition": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"computes": {
			Type:     schema.TypeList,
			Computed: true,
			Elem: &schema.Resource{
				Schema: map[string]*schema.Schema{
					"compute_id": {
						Type:     schema.TypeString,
						Computed: true,
					},
					"compute_name": {
						Type:     schema.TypeString,
						Computed: true,
					},
				},
			},
		},
		"created_time": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"deleted_time": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"destruction_time": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"devicename": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"disk_path": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"guid": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"disk_id": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"image_id": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"images": {
			Type:     schema.TypeList,
			Computed: true,
			Elem: &schema.Schema{
				Type: schema.TypeString,
			},
		},
		"iotune": {
			Type:     schema.TypeList,
			Optional: true,
			MaxItems: 1,
			Computed: true,
			Elem: &schema.Resource{
				Schema: map[string]*schema.Schema{
					"read_bytes_sec": {
						Type:     schema.TypeInt,
						Optional: true,
						Computed: true,
					},
					"read_bytes_sec_max": {
						Type:     schema.TypeInt,
						Optional: true,
						Computed: true,
					},
					"read_iops_sec": {
						Type:     schema.TypeInt,
						Optional: true,
						Computed: true,
					},
					"read_iops_sec_max": {
						Type:     schema.TypeInt,
						Optional: true,
						Computed: true,
					},
					"size_iops_sec": {
						Type:     schema.TypeInt,
						Optional: true,
						Computed: true,
					},
					"total_bytes_sec": {
						Type:     schema.TypeInt,
						Optional: true,
						Computed: true,
					},
					"total_bytes_sec_max": {
						Type:     schema.TypeInt,
						Optional: true,
						Computed: true,
					},
					"total_iops_sec": {
						Type:     schema.TypeInt,
						Optional: true,
						Computed: true,
					},
					"total_iops_sec_max": {
						Type:     schema.TypeInt,
						Optional: true,
						Computed: true,
					},
					"write_bytes_sec": {
						Type:     schema.TypeInt,
						Optional: true,
						Computed: true,
					},
					"write_bytes_sec_max": {
						Type:     schema.TypeInt,
						Optional: true,
						Computed: true,
					},
					"write_iops_sec": {
						Type:     schema.TypeInt,
						Optional: true,
						Computed: true,
					},
					"write_iops_sec_max": {
						Type:     schema.TypeInt,
						Optional: true,
						Computed: true,
					},
				},
			},
		},
		"iqn": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"login": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"milestones": {
			Type:     schema.TypeInt,
			Computed: true,
		},

		"order": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"params": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"parent_id": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"passwd": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"pci_slot": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"present_to": {
			Type:     schema.TypeList,
			Computed: true,
			Elem: &schema.Schema{
				Type: schema.TypeInt,
			},
		},
		"purge_attempts": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"purge_time": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"reality_device_number": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"reference_id": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"res_id": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"res_name": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"role": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"sep_type": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"size_used": {
			Type:     schema.TypeFloat,
			Computed: true,
		},
		"snapshots": {
			Type:     schema.TypeList,
			Computed: true,
			Elem: &schema.Resource{
				Schema: map[string]*schema.Schema{
					"guid": {
						Type:     schema.TypeString,
						Computed: true,
					},
					"label": {
						Type:     schema.TypeString,
						Computed: true,
					},
					"reference_id": {
						Type:     schema.TypeString,
						Computed: true,
					},
					"res_id": {
						Type:     schema.TypeString,
						Computed: true,
					},
					"snap_set_guid": {
						Type:     schema.TypeString,
						Computed: true,
					},
					"snap_set_time": {
						Type:     schema.TypeInt,
						Computed: true,
					},
					"timestamp": {
						Type:     schema.TypeInt,
						Computed: true,
					},
				},
			},
		},
		"status": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"tech_status": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"vmid": {
			Type:     schema.TypeInt,
			Computed: true,
		},
	}

	return rets
}

func ResourceDisk() *schema.Resource {
	return &schema.Resource{
		SchemaVersion: 1,

		CreateContext: resourceDiskCreate,
		ReadContext:   resourceDiskRead,
		UpdateContext: resourceDiskUpdate,
		DeleteContext: resourceDiskDelete,

		Importer: &schema.ResourceImporter{
			StateContext: schema.ImportStatePassthroughContext,
		},

		Timeouts: &schema.ResourceTimeout{
			Create:  &constants.Timeout180s,
			Read:    &constants.Timeout30s,
			Update:  &constants.Timeout180s,
			Delete:  &constants.Timeout60s,
			Default: &constants.Timeout60s,
		},

		Schema: resourceDiskSchemaMake(),
	}
}
