/*
Copyright (c) 2019-2022 Digital Energy Cloud Solutions LLC. All Rights Reserved.
Authors:
Petr Krutov, <petr.krutov@digitalenergy.online>
Stanislav Solovev, <spsolovev@digitalenergy.online>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
Terraform DECORT provider - manage resources provided by DECORT (Digital Energy Cloud
Orchestration Technology) with Terraform by Hashicorp.

Source code: https://repository.basistech.ru/BASIS/terraform-provider-decort

Please see README.md to learn where to place source code so that it
builds seamlessly.

Documentation: https://repository.basistech.ru/BASIS/terraform-provider-decort/wiki
*/

package rg

import (
	"context"
	"fmt"
	"strconv"

	log "github.com/sirupsen/logrus"
	"repository.basistech.ru/BASIS/decort-golang-sdk/pkg/cloudbroker/rg"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/constants"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/controller"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/dc"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/status"

	"github.com/hashicorp/terraform-plugin-sdk/v2/diag"
	"github.com/hashicorp/terraform-plugin-sdk/v2/helper/schema"
	"github.com/hashicorp/terraform-plugin-sdk/v2/helper/validation"
)

func resourceResgroupCreate(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	c := m.(*controller.ControllerCfg)

	req := rg.CreateRequest{}

	req.Name = d.Get("rg_name").(string)

	log.Debugf("resourceResgroupCreate: called for RG name %s, account ID %d",
		req.Name, d.Get("account_id").(int))

	haveAccount, err := existAccountID(ctx, d, m)
	if err != nil {
		return diag.FromErr(err)
	}
	if !haveAccount {
		return diag.Errorf("resourceResgroupCreate: can't create RG because AccountID %d is not allowed or does not exist", d.Get("account_id").(int))
	}
	req.AccountID = uint64(d.Get("account_id").(int))

	haveGID, err := existGID(ctx, d, m)
	if err != nil {
		return diag.FromErr(err)
	}
	if !haveGID {
		return diag.Errorf("resourceResgroupCreate: can't create RG because GID %d is not allowed or does not exist", d.Get("gid").(int))
	}
	req.GID = uint64(d.Get("gid").(int))

	if _, ok := d.GetOk("ext_net_id"); ok {
		haveExtNet, err := existExtNet(ctx, d, m)
		if err != nil {
			return diag.FromErr(err)
		}
		if !haveExtNet {
			return diag.Errorf("resourceResgroupCreate: can't create RG because ExtNetID %d is not allowed or does not exist", d.Get("ext_net_id").(int))
		}
	}

	if resLimits, ok := d.GetOk("resource_limits"); ok {
		resLimits := resLimits.([]interface{})[0]
		resLimitsConv := resLimits.(map[string]interface{})
		if resLimitsConv["cu_m"] != nil {
			maxMemCap := int64(resLimitsConv["cu_m"].(float64))
			if maxMemCap == 0 {
				req.MaxMemoryCapacity = -1
			} else {
				req.MaxMemoryCapacity = maxMemCap
			}
		}
		if resLimitsConv["cu_dm"] != nil {
			maxDiskCap := int64(resLimitsConv["cu_dm"].(float64))
			if maxDiskCap == 0 {
				req.MaxVDiskCapacity = -1
			} else {
				req.MaxVDiskCapacity = maxDiskCap
			}
		}
		if resLimitsConv["cu_c"] != nil {
			maxCPUCap := int64(resLimitsConv["cu_c"].(float64))
			if maxCPUCap == 0 {
				req.MaxCPUCapacity = -1
			} else {
				req.MaxCPUCapacity = maxCPUCap
			}
		}
		if resLimitsConv["cu_i"] != nil {
			maxNumPublicIP := int64(resLimitsConv["cu_i"].(float64))
			if maxNumPublicIP == 0 {
				req.MaxNumPublicIP = -1
			} else {
				req.MaxNumPublicIP = maxNumPublicIP
			}
		}
		if resLimitsConv["cu_np"] != nil {
			maxNP := int64(resLimitsConv["cu_np"].(float64))
			if maxNP == 0 {
				req.MaxNetworkPeerTransfer = -1
			} else {
				req.MaxNetworkPeerTransfer = maxNP
			}
		}
	}

	if defNetType, ok := d.GetOk("def_net_type"); ok {
		req.DefNet = defNetType.(string)
	}

	if ipcidr, ok := d.GetOk("ipcidr"); ok {
		req.IPCIDR = ipcidr.(string)
	}

	if description, ok := d.GetOk("description"); ok {
		req.Description = description.(string)
	}

	if reason, ok := d.GetOk("reason"); ok {
		req.Reason = reason.(string)
	}

	if extNetId, ok := d.GetOk("ext_net_id"); ok {
		req.ExtNetID = uint64(extNetId.(int))
	}

	if extIp, ok := d.GetOk("ext_ip"); ok {
		req.ExtIP = extIp.(string)
	}

	if regComputes, ok := d.GetOk("register_computes"); ok {
		req.RegisterComputes = regComputes.(bool)
	}

	if uniqPools, ok := d.GetOk("uniq_pools"); ok {
		uniqPools := uniqPools.([]interface{})

		for _, pool := range uniqPools {
			req.UniqPools = append(req.UniqPools, pool.(string))
		}
	}

	rgID, err := c.CloudBroker().RG().Create(ctx, req)
	if err != nil {
		return diag.FromErr(err)
	}

	d.SetId(strconv.FormatUint(rgID, 10))

	w := dc.Warnings{}

	if access, ok := d.GetOk("access"); ok {
		var user, right string

		if access.(*schema.Set).Len() > 0 {
			accessList := access.(*schema.Set).List()
			for _, accessIface := range accessList {
				access := accessIface.(map[string]interface{})
				user = access["user"].(string)
				right = access["right"].(string)

				req := rg.AccessGrantRequest{
					RGID:  rgID,
					User:  user,
					Right: right,
				}

				if reason, ok := access["reason"]; ok {
					req.Reason = reason.(string)
				}

				_, err := c.CloudBroker().RG().AccessGrant(ctx, req)
				if err != nil {
					w.Add(err)
				}
			}
		}
	}

	if defNet, ok := d.GetOk("def_net"); ok {
		if defNet.(*schema.Set).Len() > 0 {
			defNetList := defNet.(*schema.Set).List()
			defNetItem := defNetList[0].(map[string]interface{})

			netType := defNetItem["net_type"].(string)

			req := rg.SetDefNetRequest{
				RGID:    rgID,
				NetType: netType,
			}

			if netID, ok := defNetItem["net_id"]; ok {
				req.NetID = uint64(netID.(int))
			}
			if reason, ok := defNetItem["reason"]; ok {
				req.Reason = reason.(string)
			}

			_, err := c.CloudBroker().RG().SetDefNet(ctx, req)
			if err != nil {
				w.Add(err)
			}

			d.Set("def_net_type", netType)
		}

		if cpuAllocationParameter, ok := d.GetOk("cpu_allocation_parameter"); ok {
			cpuAllocationParameter := cpuAllocationParameter.(string)

			req := rg.SetCPUAllocationParameterRequest{
				RGID:        rgID,
				StrictLoose: cpuAllocationParameter,
			}

			log.Debugf("setting account cpu allocation parameter")
			_, err := c.CloudBroker().RG().SetCPUAllocationParameter(ctx, req)
			if err != nil {
				w.Add(err)
			}
		}

		if cpuAllocationRatio, ok := d.GetOk("cpu_allocation_ratio"); ok {
			cpuAllocationRatio := cpuAllocationRatio.(float64)

			req := rg.SetCPUAllocationRatioRequest{
				RGID:  rgID,
				Ratio: cpuAllocationRatio,
			}

			log.Debugf("setting account cpu allocation ratio")
			_, err := c.CloudBroker().RG().SetCPUAllocationRatio(ctx, req)
			if err != nil {
				w.Add(err)
			}
		}

		if !d.Get("enable").(bool) {
			_, err := c.CloudBroker().RG().Disable(ctx, rg.DisableRequest{
				RGID: rgID,
			})

			if err != nil {
				w.Add(err)
			}
		}
	}

	diags := resourceResgroupRead(ctx, d, m)
	diags = append(diags, w.Get()...)

	return diags
}

func resourceResgroupRead(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceResgroupRead: called for RG name %s, account ID %d",
		d.Get("rg_name").(string), d.Get("account_id").(int))

	c := m.(*controller.ControllerCfg)

	rgData, err := utilityResgroupCheckPresence(ctx, d, m)
	if err != nil {
		d.SetId("") // ensure ID is empty
		return diag.FromErr(err)
	}

	hasChanged := false

	switch rgData.Status {
	case status.Modeled:
		return diag.Errorf("The resource group is in status: %s, please, contact support for more information", rgData.Status)
	case status.Created:
	case status.Enabled:
	case status.Deleted:
		restoreReq := rg.RestoreRequest{RGID: rgData.ID}
		enableReq := rg.EnableRequest{RGID: rgData.ID}

		log.Debugf("restoring RG")
		_, err := c.CloudBroker().RG().Restore(ctx, restoreReq)
		if err != nil {
			return diag.FromErr(err)
		}

		log.Debugf("enabling RG")
		_, err = c.CloudBroker().RG().Enable(context.Background(), enableReq)
		if err != nil {
			return diag.FromErr(err)
		}

		hasChanged = true
	case status.Deleting:
	case status.Destroyed:
		d.SetId("")
		return resourceResgroupCreate(ctx, d, m)
	case status.Destroying:
	case status.Disabled:
	case status.Disabling:
	case status.Enabled:
	case status.Enabling:
	}

	if hasChanged {
		rgData, err = utilityResgroupCheckPresence(ctx, d, m)
		if err != nil {
			d.SetId("")
			return diag.FromErr(err)
		}
	}

	flattenResourceRG(d, rgData)

	return nil
}

func resourceResgroupUpdate(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceResgroupUpdate: called for RG name %s, account ID %d",
		d.Get("name").(string), d.Get("account_id").(int))

	c := m.(*controller.ControllerCfg)

	haveAccount, err := existAccountID(ctx, d, m)
	if err != nil {
		return diag.FromErr(err)
	}
	if !haveAccount {
		return diag.Errorf("resourceResgroupUpdate: can't create RG bacause AccountID %d not allowed or does not exist", d.Get("account_id").(int))
	}

	haveGID, err := existGID(ctx, d, m)
	if err != nil {
		return diag.FromErr(err)
	}
	if !haveGID {
		return diag.Errorf("resourceResgroupUpdate: can't create RG because GID %d not allowed or does not exist", d.Get("gid").(int))
	}

	if _, ok := d.GetOk("ext_net_id"); ok {
		haveExtNet, err := existExtNet(ctx, d, m)
		if err != nil {
			return diag.FromErr(err)
		}
		if !haveExtNet {
			return diag.Errorf("resourceResgroupUpdate: can't create RG bacause ExtNetID %d not allowed or does not exist", d.Get("ext_net_id").(int))
		}
	}

	rgData, err := utilityResgroupCheckPresence(ctx, d, m)
	if err != nil {
		d.SetId("")
		return diag.FromErr(err)
	}

	hasChanged := false

	switch rgData.Status {
	case status.Modeled:
	case status.Created:
	case status.Enabled:
	case status.Deleted:
		restoreReq := rg.RestoreRequest{RGID: rgData.ID}
		enableReq := rg.EnableRequest{RGID: rgData.ID}

		_, err := c.CloudBroker().RG().Restore(ctx, restoreReq)
		if err != nil {
			return diag.FromErr(err)
		}

		_, err = c.CloudBroker().RG().Enable(ctx, enableReq)
		if err != nil {
			return diag.FromErr(err)
		}

		hasChanged = true
	case status.Deleting:
	case status.Destroyed:
		d.SetId("")
		return resourceResgroupCreate(ctx, d, m)
	case status.Destroying:
	case status.Disabled:
	case status.Disabling:
	case status.Enabled:
	case status.Enabling:
	}

	if hasChanged {
		rgData, err = utilityResgroupCheckPresence(ctx, d, m)
		if err != nil {
			d.SetId("")
			return diag.FromErr(err)
		}
	}

	/* NOTE: we do not allow changing the following attributes of an existing RG via terraform:
	   - def_net_type
	   - ipcidr
	   - ext_net_id
	   - ext_ip

	   The following code fragment checks if any of these have been changed and generates error.
	*/
	if ok := d.HasChange("def_net"); ok {
		_, newDefNet := d.GetChange("def_net")
		if newDefNet.(*schema.Set).Len() == 0 {
			return diag.Errorf("resourceResgroupUpdate: block def_net must not be empty")
		}
	}

	for _, attr := range []string{"def_net_type", "ipcidr", "ext_ip"} {
		attrNew, attrOld := d.GetChange(attr)
		if attrNew.(string) != attrOld.(string) {
			return diag.FromErr(fmt.Errorf("resourceResgroupUpdate: RG ID %s: changing %s for existing RG is not allowed", d.Id(), attr))
		}
	}

	attrNew, attrOld := d.GetChange("ext_net_id")
	if attrNew.(int) != attrOld.(int) {
		return diag.FromErr(fmt.Errorf("resourceResgroupUpdate: RG ID %s: changing ext_net_id for existing RG is not allowed", d.Id()))
	}

	doGeneralUpdate := false

	req := rg.UpdateRequest{
		RGID: rgData.ID,
	}

	if d.HasChange("rg_name") {
		req.Name = d.Get("rg_name").(string)
		doGeneralUpdate = true
	}

	if d.HasChange("resource_limits") {
		resLimits := d.Get("resource_limits").([]interface{})[0]
		resLimitsConv := resLimits.(map[string]interface{})
		if resLimitsConv["cu_m"] != nil {
			maxMemCap := int64(resLimitsConv["cu_m"].(float64))
			if maxMemCap == 0 {
				req.MaxMemoryCapacity = -1
			} else {
				req.MaxMemoryCapacity = maxMemCap
			}
		}
		if resLimitsConv["cu_dm"] != nil {
			maxDiskCap := int64(resLimitsConv["cu_dm"].(float64))
			if maxDiskCap == 0 {
				req.MaxVDiskCapacity = -1
			} else {
				req.MaxVDiskCapacity = maxDiskCap
			}
		}
		if resLimitsConv["cu_c"] != nil {
			maxCPUCap := int64(resLimitsConv["cu_c"].(float64))
			if maxCPUCap == 0 {
				req.MaxCPUCapacity = -1
			} else {
				req.MaxCPUCapacity = maxCPUCap
			}
		}
		if resLimitsConv["cu_i"] != nil {
			maxNumPublicIP := int64(resLimitsConv["cu_i"].(float64))
			if maxNumPublicIP == 0 {
				req.MaxNumPublicIP = -1
			} else {
				req.MaxNumPublicIP = maxNumPublicIP
			}
		}
		if resLimitsConv["cu_np"] != nil {
			maxNP := int64(resLimitsConv["cu_np"].(float64))
			if maxNP == 0 {
				req.MaxNetworkPeerTransfer = -1
			} else {
				req.MaxNetworkPeerTransfer = maxNP
			}
		}

		doGeneralUpdate = true
	}

	if d.HasChange("desc") {
		req.Description = d.Get("desc").(string)
		doGeneralUpdate = true
	}

	if d.HasChange("register_computes") {
		req.RegisterComputes = d.Get("register_computes").(bool)
		doGeneralUpdate = true
	}

	if d.HasChange("uniq_pools") {
		uniqPools := d.Get("uniq_pools").([]interface{})
		for _, pool := range uniqPools {
			req.UniqPools = append(req.UniqPools, pool.(string))
		}

		doGeneralUpdate = true
	}

	if doGeneralUpdate {
		log.Debugf("resourceResgroupUpdate: detected delta between new and old RG specs - updating the RG")
		_, err := c.CloudBroker().RG().Update(ctx, req)
		if err != nil {
			return diag.FromErr(err)
		}
	} else {
		log.Debugf("resourceResgroupUpdate: no difference between old and new state - no update on the RG will be done")
	}

	if d.HasChange("enable") {
		enable := d.Get("enable").(bool)

		if enable && rgData.Status == status.Disabled {
			_, err := c.CloudBroker().RG().Enable(ctx, rg.EnableRequest{RGID: rgData.ID})
			if err != nil {
				return diag.FromErr(err)
			}
		} else if !enable && rgData.Status == status.Enabled {
			req := rg.DisableRequest{RGID: rgData.ID}

			_, err := c.CloudBroker().RG().Disable(ctx, req)
			if err != nil {
				return diag.FromErr(err)
			}
		}
	} //  убрать в конец

	oldSet, newSet := d.GetChange("access")

	deletedAccess := oldSet.(*schema.Set).Difference(newSet.(*schema.Set)).List()
	for _, deletedIface := range deletedAccess {
		deleteItem := deletedIface.(map[string]interface{})
		user := deleteItem["user"].(string)

		_, err := c.CloudBroker().RG().AccessRevoke(ctx, rg.AccessRevokeRequest{
			RGID: rgData.ID,
			User: user,
		})
		if err != nil {
			return diag.FromErr(err)
		}
	}

	addedAccess := newSet.(*schema.Set).Difference(oldSet.(*schema.Set)).List()
	for _, addedIface := range addedAccess {
		addedItem := addedIface.(map[string]interface{})
		user := addedItem["user"].(string)
		right := addedItem["right"].(string)

		_, err := c.CloudBroker().RG().AccessGrant(ctx, rg.AccessGrantRequest{
			RGID:  rgData.ID,
			User:  user,
			Right: right,
		})
		if err != nil {
			return diag.FromErr(err)
		}
	}

	if ok := d.HasChange("def_net"); ok {
		oldDefNet, newDefNet := d.GetChange("def_net")
		if newDefNet.(*schema.Set).Len() > 0 {
			changedDefNet := (newDefNet.(*schema.Set).Difference(oldDefNet.(*schema.Set))).List()
			for _, changedDefNetInterface := range changedDefNet {
				defNetItem := changedDefNetInterface.(map[string]interface{})
				netType := defNetItem["net_type"].(string)

				req := rg.SetDefNetRequest{
					RGID:    rgData.ID,
					NetType: netType,
				}

				if netID, ok := defNetItem["net_id"]; ok {
					req.NetID = uint64(netID.(int))
				}
				if reason, ok := defNetItem["reason"]; ok {
					req.Reason = reason.(string)
				}

				_, err := c.CloudBroker().RG().SetDefNet(ctx, req)
				if err != nil {
					return diag.FromErr(err)
				}
			}
		}
	}

	if d.HasChange("cpu_allocation_parameter") {
		cpuAllocationParameter := d.Get("cpu_allocation_parameter").(string)
		_, err := c.CloudBroker().RG().SetCPUAllocationParameter(ctx, rg.SetCPUAllocationParameterRequest{
			RGID:        rgData.ID,
			StrictLoose: cpuAllocationParameter,
		})
		if err != nil {
			return diag.FromErr(err)
		}
	}

	if d.HasChange("cpu_allocation_ratio") {
		cpuAllocationRatio := d.Get("cpu_allocation_ratio").(float64)
		_, err := c.CloudBroker().RG().SetCPUAllocationRatio(ctx, rg.SetCPUAllocationRatioRequest{
			RGID:  rgData.ID,
			Ratio: cpuAllocationRatio,
		})
		if err != nil {
			return diag.FromErr(err)
		}
	}

	return nil // убери
}

func resourceResgroupDelete(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceResgroupDelete: called for RG name %s, account ID %d",
		d.Get("rg_name").(string), d.Get("account_id").(int))

	rg_facts, err := utilityResgroupCheckPresence(ctx, d, m)
	if rg_facts == nil {
		if err != nil {
			return diag.FromErr(err)
		}
		return nil
	}

	req := rg.DeleteRequest{
		RGID:        rg_facts.ID,
		Force:       true,
		Permanently: true,
		Reason:      "Destroyed by DECORT Terraform provider",
	}

	c := m.(*controller.ControllerCfg)
	_, err = c.CloudBroker().RG().Delete(ctx, req)
	if err != nil {
		return diag.FromErr(err)
	}

	return nil
}

func ResourceResgroup() *schema.Resource {
	return &schema.Resource{
		SchemaVersion: 1,

		CreateContext: resourceResgroupCreate,
		ReadContext:   resourceResgroupRead,
		UpdateContext: resourceResgroupUpdate,
		DeleteContext: resourceResgroupDelete,

		Importer: &schema.ResourceImporter{
			StateContext: schema.ImportStatePassthroughContext,
		},

		Timeouts: &schema.ResourceTimeout{
			Create:  &constants.Timeout180s,
			Read:    &constants.Timeout600s,
			Update:  &constants.Timeout180s,
			Delete:  &constants.Timeout60s,
			Default: &constants.Timeout60s,
		},

		Schema: map[string]*schema.Schema{
			"account_id": {
				Type:         schema.TypeInt,
				Required:     true,
				ValidateFunc: validation.IntAtLeast(1),
				Description:  "Unique ID of the account, which this resource group belongs to.",
			},

			"gid": {
				Type:        schema.TypeInt,
				Required:    true,
				ForceNew:    true,
				Description: "Unique ID of the grid, where this resource group is deployed.",
			},

			"rg_name": {
				Type:        schema.TypeString,
				Required:    true,
				Description: "Name of this resource group. Names are case sensitive and unique within the context of a account.",
			},

			"resource_limits": {
				Type:     schema.TypeList,
				Optional: true,
				Computed: true,
				MaxItems: 1,
				Elem: &schema.Resource{
					Schema: map[string]*schema.Schema{
						"cu_c": {
							Type:     schema.TypeFloat,
							Optional: true,
							Computed: true,
						},
						"cu_d": {
							Type:     schema.TypeFloat,
							Computed: true,
						},
						"cu_dm": {
							Type:     schema.TypeFloat,
							Optional: true,
							Computed: true,
						},
						"cu_i": {
							Type:     schema.TypeFloat,
							Optional: true,
							Computed: true,
						},
						"cu_m": {
							Type:     schema.TypeFloat,
							Optional: true,
							Computed: true,
						},
						"cu_np": {
							Type:     schema.TypeFloat,
							Optional: true,
							Computed: true,
						},
						"gpu_units": {
							Type:     schema.TypeFloat,
							Computed: true,
						},
					},
				},
			},

			"def_net_type": {
				Type:     schema.TypeString,
				Optional: true,
				Computed: true,
				// Default:      "PRIVATE",
				ValidateFunc: validation.StringInSlice([]string{"PRIVATE", "PUBLIC", "NONE"}, false),
				Description:  "Type of the network, which this resource group will use as default for its computes - PRIVATE or PUBLIC or NONE.",
			},

			"ipcidr": {
				Type:        schema.TypeString,
				Optional:    true,
				Description: "Address of the netowrk inside the private network segment (aka ViNS) if def_net_type=PRIVATE",
			},

			"description": {
				Type:        schema.TypeString,
				Optional:    true,
				Computed:    true,
				Description: "User-defined text description of this resource group.",
			},

			"reason": {
				Type:     schema.TypeString,
				Optional: true,
			},

			"ext_net_id": {
				Type:        schema.TypeInt,
				Optional:    true,
				Default:     0,
				Description: "ID of the external network for default ViNS. Pass 0 if def_net_type=PUBLIC or no external connection required for the defult ViNS when def_net_type=PRIVATE",
			},

			"ext_ip": {
				Type:        schema.TypeString,
				Optional:    true,
				Description: "IP address on the external netowrk to request when def_net_type=PRIVATE and ext_net_id is not 0",
			},

			"register_computes": {
				Type:     schema.TypeBool,
				Optional: true,
				Computed: true,
				// Default:     false,
				Description: "Register computes in registration system",
			},

			"uniq_pools": {
				Type:     schema.TypeList,
				Computed: true,
				Optional: true,
				Elem: &schema.Schema{
					Type: schema.TypeString,
				},
			},

			"access": {
				Type:     schema.TypeSet,
				Optional: true,
				Elem: &schema.Resource{
					Schema: map[string]*schema.Schema{
						"user": {
							Type:        schema.TypeString,
							Required:    true,
							Description: "User or group name to grant access",
						},
						"right": {
							Type:        schema.TypeString,
							Required:    true,
							Description: "Access rights to set, one of 'R', 'RCX' or 'ARCXDU'",
						},
						"reason": {
							Type:        schema.TypeString,
							Optional:    true,
							Description: "Reason for action",
						},
					},
				},
			},

			"def_net": {
				Type:     schema.TypeSet,
				Optional: true,
				MaxItems: 1,
				Elem: &schema.Resource{
					Schema: map[string]*schema.Schema{
						"net_type": {
							Type:         schema.TypeString,
							Required:     true,
							ValidateFunc: validation.StringInSlice([]string{"PRIVATE", "PUBLIC"}, false),
							Description:  "Network type to set. Must be on of 'PRIVATE' or 'PUBLIC'.",
						},
						"net_id": {
							Type:        schema.TypeInt,
							Optional:    true,
							Default:     0,
							Description: "Network segment ID. If netType is PUBLIC and netId is 0 then default external network segment will be selected. If netType is PRIVATE and netId=0, the first ViNS defined for this RG will be selected. Otherwise, netId identifies either existing external network segment or ViNS.",
						},
						"reason": {
							Type:        schema.TypeString,
							Optional:    true,
							Description: "Reason for action",
						},
					},
				},
			},

			"cpu_allocation_parameter": {
				Type:        schema.TypeString,
				Optional:    true,
				Computed:    true,
				Description: "set cpu allocation parameter",
			},

			"cpu_allocation_ratio": {
				Type:        schema.TypeFloat,
				Optional:    true,
				Computed:    true,
				Description: "set cpu allocation ratio",
			},

			"enable": {
				Type:        schema.TypeBool,
				Optional:    true,
				Default:     true,
				Description: "enable/disable account",
			},

			"acl": {
				Type:     schema.TypeList,
				Computed: true,
				Elem: &schema.Resource{
					Schema: map[string]*schema.Schema{
						"explicit": {
							Type:     schema.TypeBool,
							Computed: true,
						},
						"guid": {
							Type:     schema.TypeString,
							Computed: true,
						},
						"right": {
							Type:     schema.TypeString,
							Computed: true,
						},
						"status": {
							Type:     schema.TypeString,
							Computed: true,
						},
						"type": {
							Type:     schema.TypeString,
							Computed: true,
						},
						"user_group_id": {
							Type:     schema.TypeString,
							Computed: true,
						},
					},
				},
			},

			"account_name": {
				Type:        schema.TypeString,
				Computed:    true,
				Description: "Name of the account, which this resource group belongs to.",
			},
			"created_by": {
				Type:     schema.TypeString,
				Computed: true,
			},
			"created_time": {
				Type:     schema.TypeInt,
				Computed: true,
			},

			"def_net_id": {
				Type:        schema.TypeInt,
				Computed:    true,
				Description: "ID of the default network for this resource group (if any).",
			},

			"deleted_by": {
				Type:     schema.TypeString,
				Computed: true,
			},
			"deleted_time": {
				Type:     schema.TypeInt,
				Computed: true,
			},
			"guid": {
				Type:     schema.TypeInt,
				Computed: true,
			},
			"rg_id": {
				Type:     schema.TypeInt,
				Computed: true,
			},
			"lock_status": {
				Type:     schema.TypeString,
				Computed: true,
			},
			"milestones": {
				Type:     schema.TypeInt,
				Computed: true,
			},
			"resource_types": {
				Type:     schema.TypeList,
				Computed: true,
				Elem: &schema.Schema{
					Type: schema.TypeString,
				},
			},
			"secret": {
				Type:     schema.TypeString,
				Computed: true,
			},
			"status": {
				Type:        schema.TypeString,
				Computed:    true,
				Description: "Current status of this resource group.",
			},
			"updated_by": {
				Type:     schema.TypeString,
				Computed: true,
			},
			"updated_time": {
				Type:     schema.TypeInt,
				Computed: true,
			},
			"vins": {
				Type:     schema.TypeList,
				Computed: true,
				Elem: &schema.Schema{
					Type: schema.TypeInt,
				},
				Description: "List of VINs deployed in this resource group.",
			},
		},
	}
}

func flattenResourceRG(d *schema.ResourceData, rgData *rg.RecordRG) {
	d.Set("account_id", rgData.AccountID)
	d.Set("gid", rgData.GID)
	d.Set("rg_name", rgData.Name)
	d.Set("resource_limits", flattenRgResourceLimits(rgData.ResourceLimits))
	d.Set("def_net_type", rgData.DefNetType)
	d.Set("description", rgData.Description)
	d.Set("register_computes", rgData.RegisterComputes)
	d.Set("uniq_pools", rgData.UniqPools)
	d.Set("cpu_allocation_parameter", rgData.CPUAllocationParameter)
	d.Set("cpu_allocation_ratio", rgData.CPUAllocationRatio)
	d.Set("acl", flattenRgAcl(rgData.ACL))
	d.Set("account_name", rgData.AccountName)
	d.Set("created_by", rgData.CreatedBy)
	d.Set("created_time", rgData.CreatedTime)
	d.Set("def_net_id", rgData.DefNetID)
	d.Set("deleted_by", rgData.DeletedBy)
	d.Set("deleted_time", rgData.DeletedTime)
	d.Set("guid", rgData.GUID)
	d.Set("rg_id", rgData.ID)
	d.Set("lock_status", rgData.LockStatus)
	d.Set("milestones", rgData.Milestones)
	d.Set("resource_types", rgData.ResTypes)
	d.Set("secret", rgData.Secret)
	d.Set("status", rgData.Status)
	d.Set("updated_by", rgData.UpdatedBy)
	d.Set("updated_time", rgData.UpdatedTime)
	d.Set("vins", rgData.VINS)
}
